#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import JAN, MAR, APR, MAY, JUN, AUG, SEP, OCT, NOV, MON, TUE, FRI
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Australia(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
      - https://www.qld.gov.au/recreation/travel/holidays
    """

    country = "AU"
    special_holidays = {
        2022: (SEP, 22, "National Day of Mourning for Queen Elizabeth II"),
    }
    subdivisions = ("ACT", "NSW", "NT", "QLD", "SA", "TAS", "VIC", "WA")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date, include_sat: bool = True, days: int = +1) -> None:
        if not self.observed:
            return None
        if self._is_sunday(dt) or (include_sat and self._is_saturday(dt)):
            self._add_holiday(
                "%s (Observed)" % self[dt], dt + td(days=+2 if self._is_saturday(dt) else days)
            )

    def _populate(self, year):
        super()._populate(year)

        # ACT:  Holidays Act 1958
        # NSW:  Public Holidays Act 2010
        # NT:   Public Holidays Act 2013
        # QLD:  Holidays Act 1983
        # SA:   Holidays Act 1910
        # TAS:  Statutory Holidays Act 2000
        # VIC:  Public Holidays Act 1993
        # WA:   Public and Bank Holidays Act 1972

        # TODO do more research on history of Aus holidays

        # New Year's Day
        self._add_observed(self._add_new_years_day("New Year's Day"))

        # Easter
        self._add_good_friday("Good Friday")
        self._add_easter_monday("Easter Monday")

        # Sovereign's Birthday
        if 1902 <= year <= 1935:
            name = "King's Birthday"
            if self._year >= 1912:
                self._add_holiday(name, JUN, 3)  # George V
            else:
                self._add_holiday(name, NOV, 9)  # Edward VII

    # Sovereign's Birthday
    def _add_sovereign_birthday(self, dt: date) -> None:
        if self._year >= 1936:
            name = "Queen's Birthday" if 1952 <= self._year <= 2022 else "King's Birthday"
            self._add_holiday(name, dt)

    def _add_subdiv_holidays(self):
        # Australia Day
        if self._year >= 1935:
            name = (
                "Anniversary Day"
                if self.subdiv == "NSW" and self._year <= 1945
                else "Australia Day"
            )
            jan_26 = self._add_holiday(name, JAN, 26)
            if self._year >= 1946:
                self._add_observed(jan_26)
        elif self._year >= 1888 and self.subdiv != "SA":
            self._add_holiday("Anniversary Day", JAN, 26)

        # Anzac Day
        if self._year >= 1921:
            self._add_holiday("Anzac Day", APR, 25)

        # Christmas Day
        self._add_observed(self._add_christmas_day("Christmas Day"), days=+2)

        # Boxing Day
        name = "Proclamation Day" if self.subdiv == "SA" else "Boxing Day"
        self._add_observed(self._add_christmas_day_two(name), days=+2)

        super()._add_subdiv_holidays()

    def _add_subdiv_act_holidays(self):
        # Easter
        self._add_holy_saturday("Easter Saturday")
        self._add_easter_sunday("Easter Sunday")

        # Labour Day
        self._add_holiday("Labour Day", self._get_nth_weekday_of_month(1, MON, OCT))

        # Sovereign's Birthday
        self._add_sovereign_birthday(self._get_nth_weekday_of_month(2, MON, JUN))

        # Anzac Day
        if self._year >= 1921:
            self._add_observed(date(self._year, APR, 25), include_sat=False)

        # Canberra Day
        # Info from https://www.timeanddate.com/holidays/australia/canberra-day
        # and https://en.wikipedia.org/wiki/Canberra_Day
        if self._year >= 1913:
            if self._year <= 1957:
                dt = date(self._year, MAR, 12)
            elif self._year <= 2007:
                dt = self._get_nth_weekday_of_month(3, MON, MAR)
            elif self._year == 2012:
                dt = date(self._year, MAR, 12)
            else:
                dt = self._get_nth_weekday_of_month(2, MON, MAR)
            self._add_holiday("Canberra Day", dt)

        # Family & Community Day
        if 2007 <= self._year <= 2017:
            # first Monday of the September/October school holidays
            # moved to the second Monday if this falls on Labour day
            # TODO need a formula for the ACT school holidays then
            # http://www.cmd.act.gov.au/communication/holidays
            fc_dates = {
                2010: date(2010, SEP, 26),
                2011: date(2011, OCT, 10),
                2012: date(2012, OCT, 8),
                2013: date(2013, SEP, 30),
                2014: date(2014, SEP, 29),
                2015: date(2015, SEP, 28),
                2016: date(2016, SEP, 26),
                2017: date(2017, SEP, 25),
            }
            dt = fc_dates.get(self._year, self._get_nth_weekday_of_month(1, TUE, NOV))
            self._add_holiday("Family & Community Day", dt)

        # Reconciliation Day
        if self._year >= 2018:
            self._add_holiday("Reconciliation Day", self._get_nth_weekday_from(1, MON, MAY, 27))

    def _add_subdiv_nsw_holidays(self):
        # Easter
        self._add_holy_saturday("Easter Saturday")
        self._add_easter_sunday("Easter Sunday")

        # Labour Day
        self._add_holiday("Labour Day", self._get_nth_weekday_of_month(1, MON, OCT))

        # Sovereign's Birthday
        self._add_sovereign_birthday(self._get_nth_weekday_of_month(2, MON, JUN))

        # Bank Holiday
        if self._year >= 1912:
            self._add_holiday("Bank Holiday", self._get_nth_weekday_of_month(1, MON, AUG))

    def _add_subdiv_nt_holidays(self):
        # Easter
        self._add_holy_saturday("Easter Saturday")

        # Labour Day
        self._add_holiday("May Day", self._get_nth_weekday_of_month(1, MON, MAY))

        # Sovereign's Birthday
        self._add_sovereign_birthday(self._get_nth_weekday_of_month(2, MON, JUN))

        # Anzac Day
        if self._year >= 1921:
            self._add_observed(date(self._year, APR, 25))

        # Picnic Day
        self._add_holiday("Picnic Day", self._get_nth_weekday_of_month(1, MON, AUG))

    def _add_subdiv_qld_holidays(self):
        # Easter
        self._add_holy_saturday("Easter Saturday")
        self._add_easter_sunday("Easter Sunday")

        # Labour Day
        dt = (
            self._get_nth_weekday_of_month(1, MON, OCT)
            if 2013 <= self._year <= 2015
            else self._get_nth_weekday_of_month(1, MON, MAY)
        )
        self._add_holiday("Labour Day", dt)

        # Sovereign's Birthday
        if self._year == 2012:
            self._add_holiday("Queen's Diamond Jubilee", JUN, 11)

        dt = (
            self._get_nth_weekday_of_month(2, MON, JUN)
            if self._year <= 2015 and self._year != 2012
            else self._get_nth_weekday_of_month(1, MON, OCT)
        )
        self._add_sovereign_birthday(dt)

        # Anzac Day
        if self._year >= 1921:
            self._add_observed(date(self._year, APR, 25), include_sat=False)

        # The Royal Queensland Show (Ekka)
        # The Show starts on the first Friday of August - providing this is
        # not prior to the 5th - in which case it will begin on the second
        # Friday. The Wednesday during the show is a public holiday.
        ekka_dates = {
            2020: date(2020, AUG, 14),
            2021: date(2021, OCT, 29),
        }
        dt = ekka_dates.get(self._year, self._get_nth_weekday_from(1, FRI, AUG, 5) + td(days=+5))
        self._add_holiday("The Royal Queensland Show", dt)

    def _add_subdiv_sa_holidays(self):
        # Easter
        self._add_holy_saturday("Easter Saturday")

        # Labour Day
        self._add_holiday("Labour Day", self._get_nth_weekday_of_month(1, MON, OCT))

        # Sovereign's Birthday
        self._add_sovereign_birthday(self._get_nth_weekday_of_month(2, MON, JUN))

        # Anzac Day
        if self._year >= 1921:
            self._add_observed(date(self._year, APR, 25), include_sat=False)

        # Adelaide Cup
        dt = self._get_nth_weekday_of_month(2 if self._year >= 2006 else 3, MON, MAR)
        self._add_holiday("Adelaide Cup", dt)

    def _add_subdiv_tas_holidays(self):
        # Labour Day
        self._add_holiday("Eight Hours Day", self._get_nth_weekday_of_month(2, MON, MAR))

        # Sovereign's Birthday
        self._add_sovereign_birthday(self._get_nth_weekday_of_month(2, MON, JUN))

    def _add_subdiv_vic_holidays(self):
        # Easter
        self._add_holy_saturday("Easter Saturday")
        self._add_easter_sunday("Easter Sunday")

        # Labour Day
        self._add_holiday("Labour Day", self._get_nth_weekday_of_month(2, MON, MAR))

        # Sovereign's Birthday
        self._add_sovereign_birthday(self._get_nth_weekday_of_month(2, MON, JUN))

        # Melbourne Cup
        self._add_holiday("Melbourne Cup", self._get_nth_weekday_of_month(1, TUE, NOV))

        if self._year >= 2015:
            # Grand Final Day
            grand_final_dates = {
                # Rescheduled due to COVID-19
                2020: date(2020, OCT, 23),
                # Rescheduled due to COVID-19
                2021: date(2021, SEP, 24),
                2022: date(2022, SEP, 23),
            }
            dt = grand_final_dates.get(self._year, self._get_nth_weekday_from(1, FRI, SEP, 24))
            self._add_holiday("Grand Final Day", dt)

    def _add_subdiv_wa_holidays(self):
        # Labour Day
        self._add_holiday("Labour Day", self._get_nth_weekday_of_month(1, MON, MAR))

        # Sovereign's Birthday
        self._add_sovereign_birthday(self._get_nth_weekday_from(-1, MON, OCT, 1))

        # Anzac Day
        if self._year >= 1921:
            self._add_observed(date(self._year, APR, 25))

        # Western Australia Day
        if self._year >= 1833:
            name = "Western Australia Day" if self._year >= 2015 else "Foundation Day"
            self._add_holiday(name, self._get_nth_weekday_of_month(1, MON, JUN))


class AU(Australia):
    pass


class AUS(Australia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import GREGORIAN_CALENDAR, MAR, MAY, SEP, NOV
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.constants import PUBLIC, SCHOOL
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Bulgaria(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Official holidays in Bulgaria in their current form. This class does not
    any return holidays before 1990, as holidays in the People's Republic of
    Bulgaria and earlier were different.

    Since 2017, it has been accepted that public holidays in Bulgaria that fall on a Saturday
    or Sunday are to be taken on the first working day after them. If there are both Saturday
    and Sunday holidays, Monday and Tuesday are rested respectively.
    The exceptions are:
    1) the Easter holidays, which are always a consecutive Friday, Saturday, and Sunday;
    2) National Awakening Day which, while an official holiday and a non-attendance day for
    schools, is still a working day.

    Sources (Bulgarian):
    - http://lex.bg/laws/ldoc/1594373121
    - https://www.parliament.bg/bg/24
    - https://kik-info.com/spravochnik/calendar/2021/

    Sources (English):
    - https://en.wikipedia.org/wiki/Public_holidays_in_Bulgaria
    """

    country = "BG"
    default_language = "bg"
    supported_categories = {PUBLIC, SCHOOL}
    supported_languages = ("bg", "en_US", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        if self._year <= 1989:
            return None

        observed_dates = set()

        # New Year's Day.
        observed_dates.add(self._add_new_years_day(tr("Нова година")))

        observed_dates.add(
            # Liberation Day.
            self._add_holiday(tr("Ден на Освобождението на България от османско иго"), MAR, 3)
        )

        # Good Friday.
        self._add_good_friday(tr("Велики петък"))

        # Holy Saturday.
        self._add_holy_saturday(tr("Велика събота"))

        # Easter.
        name = tr("Великден")
        self._add_easter_sunday(name)
        self._add_easter_monday(name)

        observed_dates.add(
            # International Workers' Day.
            self._add_labor_day(tr("Ден на труда и на международната работническа солидарност"))
        )

        observed_dates.add(
            # Saint George's Day.
            self._add_holiday(tr("Гергьовден, Ден на храбростта и Българската армия"), MAY, 6)
        )

        observed_dates.add(
            self._add_holiday(
                # Bulgarian Education and Culture and Slavonic Literature Day.
                tr(
                    "Ден на светите братя Кирил и Методий, на българската азбука, "
                    "просвета и култура и на славянската книжовност"
                ),
                MAY,
                24,
            )
        )

        # Unification Day.
        observed_dates.add(self._add_holiday(tr("Ден на Съединението"), SEP, 6))

        # Independence Day.
        observed_dates.add(self._add_holiday(tr("Ден на Независимостта на България"), SEP, 22))

        # Christmas Eve.
        observed_dates.add(self._add_christmas_eve(tr("Бъдни вечер"), GREGORIAN_CALENDAR))

        # Christmas Day.
        name = tr("Рождество Христово")
        observed_dates.add(self._add_christmas_day(name, GREGORIAN_CALENDAR))
        observed_dates.add(self._add_christmas_day_two(name, GREGORIAN_CALENDAR))

        if self.observed and self._year >= 2017:
            excluded_names = {self.tr("Велика събота"), self.tr("Великден")}
            for dt in sorted(observed_dates):
                if not self._is_weekend(dt):
                    continue
                obs_date = dt + td(days=+2 if self._is_saturday(dt) else +1)
                while obs_date in self:
                    obs_date += td(days=+1)
                for name in self.get_list(dt):
                    if name not in excluded_names:
                        self._add_holiday(self.tr("%s (почивен ден)") % name, obs_date)

    def _populate_school_holidays(self):
        # National Awakening Day.
        self._add_holiday(tr("Ден на народните будители"), NOV, 1)


class BG(Bulgaria):
    pass


class BLG(Bulgaria):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from typing import Optional

from holidays.calendars.gregorian import JUN, JUL, AUG, SEP, OCT, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import (
    ChineseCalendarHolidays,
    ChristianHolidays,
    InternationalHolidays,
)


class HongKong(HolidayBase, ChineseCalendarHolidays, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Hong_Kong
    Holidays for 2007–2023 (government source):
    https://www.gov.hk/en/about/abouthk/holiday/index.htm
    """

    country = "HK"
    special_holidays = {
        1997: (JUL, 2, "Hong Kong Special Administrative Region Establishment Day"),
        2015: (
            (
                SEP,
                3,
                "The 70th anniversary day of the victory of the Chinese "
                "people's war of resistance against Japanese aggression",
            )
        ),
    }

    def __init__(self, *args, **kwargs):
        ChineseCalendarHolidays.__init__(self)
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_holiday(self, *args) -> Optional[date]:
        name, dt = self._parse_holiday(*args)
        if self.observed and (self._is_sunday(dt) or dt in self):
            while self._is_sunday(dt) or dt in self:
                dt += td(days=+1)
            name = self.tr("The day following %s") % self.tr(name)
        self[dt] = self.tr(name)
        return dt

    def _populate(self, year):
        # Current set of holidays actually valid since 1946
        if year <= 1945:
            return None

        super()._populate(year)

        day_following = "The day following "

        # The first day of January
        self._add_new_years_day("The first day of January")

        # Lunar New Year
        name = "Lunar New Year's Day"
        preceding_day_lunar = "The day preceding Lunar New Year's Day"
        second_day_lunar = "The second day of Lunar New Year"
        third_day_lunar = "The third day of Lunar New Year"
        fourth_day_lunar = "The fourth day of Lunar New Year"
        new_year_date = self._chinese_new_year
        if self.observed:
            if self._is_sunday(new_year_date):
                if year in {2006, 2007, 2010}:
                    self._add_chinese_new_years_eve(preceding_day_lunar)
                else:
                    self._add_chinese_new_years_day_four(fourth_day_lunar)
            else:
                self._add_chinese_new_years_day(name)
            if self._is_saturday(new_year_date):
                self._add_chinese_new_years_day_four(fourth_day_lunar)
            else:
                self._add_chinese_new_years_day_two(second_day_lunar)
            if self._is_friday(new_year_date):
                self._add_chinese_new_years_day_four(fourth_day_lunar)
            else:
                self._add_chinese_new_years_day_three(third_day_lunar)
        else:
            self._add_chinese_new_years_day(name)
            self._add_chinese_new_years_day_two(second_day_lunar)
            self._add_chinese_new_years_day_three(third_day_lunar)

        easter_monday_date = self._easter_sunday + td(days=+1)

        # Ching Ming Festival
        name = "Ching Ming Festival"
        ching_ming_date = self._qingming_date
        if self.observed and ching_ming_date == easter_monday_date:
            self._add_holiday(f"{day_following}{name}", ching_ming_date + td(days=+1))
        elif ching_ming_date not in {
            self._easter_sunday + td(days=-2),
            self._easter_sunday + td(days=-1),
        }:
            self._add_holiday(name, ching_ming_date)

        # Easter Holiday
        good_friday = "Good Friday"
        easter_monday = "Easter Monday"
        self._add_good_friday(good_friday)
        self._add_holy_saturday(f"{day_following}{good_friday}")
        self._add_easter_monday(easter_monday)

        if ching_ming_date in {
            self._easter_sunday + td(days=-2),
            self._easter_sunday + td(days=-1),
        }:
            super()._add_holiday(name, ching_ming_date)

        if year >= 1999:
            # The Birthday of the Buddha
            self._add_chinese_birthday_of_buddha("The Birthday of the Buddha")

            # Labour Day
            self._add_labor_day("Labour Day")

        # Tuen Ng Festival
        self._add_dragon_boat_festival("Tuen Ng Festival")

        # Hong Kong Special Administrative Region Establishment Day
        if year >= 1997:
            self._add_holiday("Hong Kong Special Administrative Region Establishment Day", JUL, 1)

        # Chinese Mid-Autumn Festival
        name = "Chinese Mid-Autumn Festival"
        mid_autumn_date = self._mid_autumn_festival
        if self.observed:
            # if Chinese Mid-Autumn Festival lies on Saturday
            # before 1983 public holiday lies on Monday
            # from 1983 to 2010 public holiday lies on same day
            # since 2011 public holiday lies on Monday
            if self._is_saturday(mid_autumn_date):
                if 1983 <= year <= 2010:
                    self._add_mid_autumn_festival(name)
                else:
                    self._add_holiday(
                        f"The second day of the {name} (Monday)", mid_autumn_date + td(days=+2)
                    )
            else:
                self._add_mid_autumn_festival_day_two(f"The day following the {name}")
        else:
            self._add_mid_autumn_festival_day_two(name)

        # Chung Yeung Festival
        self._add_double_ninth_festival("Chung Yeung Festival")

        # National Day
        if year >= 1997:
            self._add_holiday("National Day", OCT, 1)

            if year <= 1998:
                self._add_holiday("National Day", OCT, 2)

        # Christmas Day
        name = "Christmas Day"
        first_after_christmas = f"The first weekday after {name}"
        second_after_christmas = f"The second weekday after {name}"
        christmas_date = self._christmas_day
        if self.observed:
            if self._is_sunday(christmas_date):
                self._add_christmas_day_two(first_after_christmas)
                self._add_christmas_day_three(second_after_christmas)
            elif self._is_saturday(christmas_date):
                self._add_christmas_day(name)
                self._add_christmas_day_three(first_after_christmas)
            else:
                self._add_christmas_day(name)
                self._add_christmas_day_two(first_after_christmas)
        else:
            self._add_christmas_day(name)
            self._add_christmas_day_two(first_after_christmas)

        # Previous holidays
        if 1952 <= year <= 1997:
            # Queen's Birthday (June 2nd Monday)
            self._add_holiday("Queen's Birthday", self._get_nth_weekday_of_month(2, MON, JUN))

        if year <= 1996:
            # Anniversary of the liberation of Hong Kong (August last Monday)
            self._add_holiday(
                "Anniversary of the liberation of Hong Kong",
                self._get_nth_weekday_of_month(-1, MON, AUG),
            )

        if year <= 1998:
            # Anniversary of the victory in the Second Sino-Japanese War
            super()._add_holiday(
                "Anniversary of the victory in the Second Sino-Japanese War",
                self._get_nth_weekday_of_month(-1, MON, AUG) + td(days=-1),
            )


class HK(HongKong):
    pass


class HKG(HongKong):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr
from typing import Set

from holidays.calendars.gregorian import JAN, FEB, SEP, NOV, THU, FRI, SAT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import IslamicHolidays


class SaudiArabia(HolidayBase, IslamicHolidays):
    """
    There are only 4 official national holidays in Saudi:
    https://laboreducation.hrsd.gov.sa/en/gallery/274
    https://laboreducation.hrsd.gov.sa/en/labor-education/322
    https://english.alarabiya.net/News/gulf/2022/01/27/Saudi-Arabia-to-commemorate-Founding-Day-on-Feb-22-annually-Royal-order
    The national day and the founding day holidays are based on the
    Georgian calendar while the other two holidays are based on the
    Islamic Calendar, and they are estimates as they announced each
    year and based on moon sightings;
    they are:
        - Eid al-Fitr
        - Eid al-Adha
    """

    country = "SA"
    default_language = "ar"
    # Estimated label.
    estimated_label = tr("(تقدير*) *%s")
    supported_languages = ("ar", "en_US")

    special_holidays = {
        # Celebrate the country's win against Argentina in the World Cup
        2022: (NOV, 23, tr("يوم وطني")),
    }

    def __init__(self, *args, **kwargs):
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_islamic_observed(self, hol_name: str, hol_dates: Set[date]) -> None:
        if not self.observed:
            return None
        for dt in hol_dates:
            weekend_days = sum(self._is_weekend(dt + td(days=i)) for i in range(4))
            for i in range(weekend_days):
                self._add_holiday(self.tr("(ملاحظة) %s") % self.tr(hol_name), dt + td(days=+4 + i))

    def _add_observed(self, hol_date: date) -> None:
        if self.observed:
            weekend = sorted(self.weekend)
            # 1st weekend day (Thursday before 2013 and Friday otherwise)
            if hol_date.weekday() == weekend[0]:
                self._add_holiday(
                    self.tr("(ملاحظة) %s") % self.tr(self[hol_date]), hol_date + td(days=-1)
                )
            # 2nd weekend day (Friday before 2013 and Saturday otherwise)
            elif hol_date.weekday() == weekend[1]:
                self._add_holiday(
                    self.tr("(ملاحظة) %s") % self.tr(self[hol_date]), hol_date + td(days=+1)
                )

    def _populate(self, year):
        super()._populate(year)

        # Weekend used to be THU, FRI before June 28th, 2013.
        # On that year both Eids were after that date, and Founding day
        # holiday started at 2022; so what below works.
        self.weekend = {THU, FRI} if year <= 2012 else {FRI, SAT}

        # The holiday is a 4-day holiday starting on the day following the
        # 29th day of Ramadan, the 9th month of the Islamic calendar.
        # Observed days are added to make up for any days falling on a weekend.
        # Holidays may straddle across Gregorian years, so we go back one year
        # to pick up any such occurrence.
        # Date of observance is announced yearly.
        # Eid al-Fitr Holiday
        name = tr("عطلة عيد الفطر")
        dates = self._add_eid_al_fitr_day(name)
        self._add_eid_al_fitr_day_two(name)
        self._add_eid_al_fitr_day_three(name)
        self._add_eid_al_fitr_day_four(name)
        self._add_islamic_observed(name, dates)

        # The holiday is a 4-day holiday starting on Arafat Day, the 10th of
        # Dhu al-Hijjah, the 12th month of the Islamic calendar.
        # Observed days are added to make up for any days falling on a weekend.
        # Holidays may straddle across Gregorian years, so we go back one year
        # to pick up any such occurrence.
        # Date of observance is announced yearly.
        # Arafat Day
        dates = self._add_arafah_day(tr("يوم عرفة"))
        # Eid al-Adha Holiday
        name = tr("عطلة عيد الأضحى")
        self._add_eid_al_adha_day(name)
        self._add_eid_al_adha_day_two(name)
        self._add_eid_al_adha_day_three(name)
        self._add_islamic_observed(name, dates)

        # National Day holiday (started at the year 2005).
        # Note: if national day happens within the Eid al-Fitr Holiday or
        # within Eid al-Fitr Holiday, there is no extra holidays given for it.
        if year >= 2005:
            dt = date(year, SEP, 23)
            if dt not in self:
                # National Day Holiday
                self._add_observed(self._add_holiday(tr("اليوم الوطني"), dt))

        # Founding Day holiday (started 2022).
        # Note: if founding day happens within the Eid al-Fitr Holiday or
        # within Eid al-Fitr Holiday, there is no extra holidays given for it.
        if year >= 2022:
            dt = date(year, FEB, 22)
            if dt not in self:
                # Founding Day
                self._add_observed(self._add_holiday(tr("يوم التأسيسي"), dt))

        # observed holidays special cases
        if self.observed and year == 2001:
            # Eid al-Fitr Holiday (observed)
            self._add_holiday(tr("عطلة عيد الفطر (ملاحظة)"), JAN, 1)


class SA(SaudiArabia):
    pass


class SAU(SaudiArabia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import FEB
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Serbia(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Serbia holidays.

    References:
     - https://en.wikipedia.org/wiki/Public_holidays_in_Serbia
    """

    country = "RS"
    default_language = "sr"
    supported_languages = ("en_US", "sr")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date, days: int = +1) -> None:
        if self.observed and self._is_sunday(dt):
            self._add_holiday(self.tr("%s (слободан дан)") % self[dt], dt + td(days))

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        name = tr("Нова година")
        self._add_observed(self._add_new_years_day(name), days=+2)
        self._add_observed(self._add_new_years_day_two(name))

        # Orthodox Christmas.
        self._add_christmas_day(tr("Божић"))

        # Statehood Day.
        name = tr("Дан државности Србије")
        self._add_observed(self._add_holiday(name, FEB, 15), days=+2)
        self._add_observed(self._add_holiday(name, FEB, 16))

        # International Workers' Day.
        name = tr("Празник рада")
        self._add_observed(self._add_labor_day(name), days=+2)
        may_2 = self._add_labor_day_two(name)
        self._add_observed(may_2, days=+2 if may_2 == self._easter_sunday else +1)

        # Armistice Day.
        self._add_observed(self._add_remembrance_day(tr("Дан примирја у Првом светском рату")))

        # Good Friday.
        self._add_good_friday(tr("Велики петак"))
        # Easter Saturday.
        self._add_holy_saturday(tr("Велика субота"))
        # Easter Sunday.
        self._add_easter_sunday(tr("Васкрс"))
        # Easter Monday.
        self._add_easter_monday(tr("Други дан Васкрса"))


class RS(Serbia):
    pass


class SRB(Serbia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from gettext import gettext as tr

from holidays.calendars.gregorian import FEB, MAR, SEP, NOV, DEC, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Mexico(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
    - https://en.wikipedia.org/wiki/Public_holidays_in_Mexico
    - https://es.wikipedia.org/wiki/Anexo:D%C3%ADas_festivos_en_M%C3%A9xico
    - https://www.gob.mx/cms/uploads/attachment/file/156203/1044_Ley_Federal_del_Trabajo.pdf
    - http://www.diputados.gob.mx/LeyesBiblio/ref/lft/LFT_orig_01abr70_ima.pdf
    """

    country = "MX"
    default_language = "es"
    supported_languages = ("en_US", "es", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Año Nuevo"))

        if year >= 1917:
            self._add_holiday(
                # Constitution Day.
                tr("Día de la Constitución"),
                self._get_nth_weekday_of_month(1, MON, FEB)
                if year >= 2006
                else date(year, FEB, 5),
            )

        if year >= 1917:
            self._add_holiday(
                # Benito Juárez's birthday.
                tr("Natalicio de Benito Juárez"),
                self._get_nth_weekday_of_month(3, MON, MAR)
                # no 2006 due to celebration of the 200th anniversary
                # of Benito Juárez in 2006
                if year >= 2007 else date(year, MAR, 21),
            )

        if year >= 1923:
            # Labor Day.
            self._add_labor_day(tr("Día del Trabajo"))

        # Independence Day.
        self._add_holiday(tr("Día de la Independencia"), SEP, 16)

        if year >= 1917:
            self._add_holiday(
                # Revolution Day.
                tr("Día de la Revolución"),
                self._get_nth_weekday_of_month(3, MON, NOV)
                if year >= 2006
                else date(year, NOV, 20),
            )

        if year >= 1970 and (year - 1970) % 6 == 0:
            # Change of Federal Government.
            self._add_holiday(tr("Transmisión del Poder Ejecutivo Federal"), DEC, 1)

        # Christmas Day.
        self._add_christmas_day(tr("Navidad"))


class MX(Mexico):
    pass


class MEX(Mexico):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, MAR, APR, AUG
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Aruba(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://www.government.aw/information-public-services/hiring-people_47940/item/holidays_43823.html  # noqa: E501
    https://www.overheid.aw/informatie-dienstverlening/ondernemen-en-werken-subthemas_46970/item/feestdagen_37375.html  # noqa: E501
    https://www.gobierno.aw/informacion-tocante-servicio/haci-negoshi-y-traha-sub-topics_47789/item/dia-di-fiesta_41242.html  # noqa: E501
    https://www.visitaruba.com/about-aruba/national-holidays-and-celebrations/
    https://www.arubatoday.com/we-celebrate-our-national-hero-betico-croes/
    https://www.caribbeannewsglobal.com/carnival-monday-remains-a-festive-day-in-aruba/  # noqa: E501
    https://www.aruba.com/us/calendar/national-anthem-and-flag-day
    """

    country = "AW"
    default_language = "pap"
    supported_languages = ("en_US", "nl", "pap", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        # AUG 1947: Autonomous State status in the Kingdom of the Netherlands.
        if year <= 1946:
            return None

        super()._populate(year)

        # Aña Nobo.
        # Status: In-Use.

        # New Year's Day
        self._add_new_years_day(tr("Aña Nobo"))

        # Dia Di Betico.
        # Status: In-Use.
        # Started in 1989.

        if year >= 1989:
            # Betico Day
            self._add_holiday(tr("Dia di Betico"), JAN, 25)

        # Dialuna prome cu diaranson di shinish.
        # Status: In-Use.
        # Starts as a public holiday from 1956 onwards.
        # Event cancelled but remain a holiday in 2021.
        # Have its name changed from 2023 onwards.

        if year >= 1956:
            self._add_ash_monday(
                # Carnival Monday
                tr("Dialuna despues di Carnaval Grandi")
                if year <= 2022
                # Monday before Ash Wednesday
                else tr("Dialuna prome cu diaranson di shinish")
            )

        # Dia di Himno y Bandera.
        # Status: In-Use.
        # Started in 1976.

        if year >= 1976:
            # National Anthem and Flag Day
            self._add_holiday(tr("Dia di Himno y Bandera"), MAR, 18)

        # Bierna Santo.
        # Status: In-Use.

        # Good Friday
        self._add_good_friday(tr("Bierna Santo"))

        # Di dos dia di Pasco di Resureccion.
        # Status: In-Use.

        # Easter Monday
        self._add_easter_monday(tr("Di dos dia di Pasco di Resureccion"))

        # Aña di La Reina/Aña di Rey/Dia di Rey.
        # Status: In-Use.
        # Started under Queen Wilhelmina in 1891.
        # Queen Beatrix kept Queen Juliana's Birthday after her coronation.
        # Switched to Aña di Rey in 2014 for King Willem-Alexander.
        # Have its name changed again to Dia di Rey from 2021 onwards.

        # King's / Queen's Day
        name = (
            # King's Day.
            tr("Dia di Rey")
            if year >= 2021
            else (
                # King's Day.
                tr("Aña di Rey")
                if year >= 2014
                # Queen's Day.
                else tr("Aña di La Reina")
            )
        )
        if year >= 2014:
            dt = date(year, APR, 27)
        elif year >= 1949:
            dt = date(year, APR, 30)
        else:
            dt = date(year, AUG, 31)
        if self._is_sunday(dt):
            dt += td(days=-1) if year >= 1980 else td(days=+1)
        self._add_holiday(name, dt)

        # Dia di Labor/Dia di Obrero.
        # Status: In-Use.

        # Labor Day
        self._add_labor_day(tr("Dia di Obrero"))

        # Dia di Asuncion.
        # Status: In-Use.

        # Ascension Day
        self._add_ascension_thursday(tr("Dia di Asuncion"))

        # Pasco di Nacemento.
        # Status: In-Use.

        # Christmas Day
        self._add_christmas_day(tr("Pasco di Nacemento"))

        # Di dos dia di Pasco di Nacemento.
        # Status: In-Use.

        # Second Day of Christmas
        self._add_christmas_day_two(tr("Di dos dia di Pasco di Nacemento"))


class AW(Aruba):
    pass


class ABW(Aruba):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import FEB, MAR, APR, MAY, AUG, NOV
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, IslamicHolidays, InternationalHolidays


class Kyrgyzstan(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """
    Kyrgyzstan holidays.

    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_Kyrgyzstan
    """

    country = "KG"

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day("New Year's Day")

        # Orthodox Christmas.
        self._add_christmas_day("Christmas Day")

        # Feb. 23 Fatherland Defender's Day.
        self._add_holiday("Fatherland Defender's Day", FEB, 23)

        # International Women's Day.
        self._add_womens_day("International Women's Day")

        # Nooruz Mairamy.
        self._add_holiday("Nooruz Mairamy", MAR, 21)

        if year >= 2016:
            # Day of the People's April Revolution.
            self._add_holiday("Day of the People's April Revolution", APR, 7)

        # International Workers' Day.
        self._add_labor_day("International Workers' Day")

        # Constitution Day.
        self._add_holiday("Constitution Day", MAY, 5)

        # Victory Day.
        self._add_world_war_two_victory_day("Victory Day")

        # Independence Day.
        self._add_holiday("Independence Day", AUG, 31)

        # Days History and Commemoration of Ancestors.
        name = "Days of History and Commemoration of Ancestors"
        self._add_holiday(name, NOV, 7)
        self._add_holiday(name, NOV, 8)

        # New Year's Eve.
        self._add_new_years_eve("New Year's Eve")

        # Islamic Holidays.
        # Orozo Ait.
        name = "Orozo Ait"
        self._add_eid_al_fitr_day(name)
        self._add_eid_al_fitr_day_two(name)

        # Kurman Ait.
        self._add_eid_al_adha_day("Kurman Ait")


class KG(Kyrgyzstan):
    pass


class KGZ(Kyrgyzstan):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import APR, MAY, JUL, AUG, OCT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import IslamicHolidays, InternationalHolidays


class Turkey(HolidayBase, IslamicHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Turkey
    """

    country = "TR"

    def __init__(self, *args, **kwargs):
        IslamicHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # 1st of Jan.
        self._add_new_years_day("New Year's Day")

        # 23rd of Apr.
        self._add_holiday("National Sovereignty and Children's Day", APR, 23)

        # 1st of May.
        self._add_labor_day("Labour Day")

        # 19th of May.
        self._add_holiday("Commemoration of Ataturk, Youth and Sports Day", MAY, 19)

        # 15th of Jul.
        # Became a national holiday after 15 Jul 2016 coup d'etat attempt.
        if year > 2016:
            self._add_holiday("Democracy and National Unity Day", JUL, 15)

        # 30th of Aug.
        self._add_holiday("Victory Day", AUG, 30)

        # 29th of Oct.
        self._add_holiday("Republic Day", OCT, 29)

        # Ramadan Feast.
        # Date of observance is announced yearly. This is an estimate.
        self._add_eid_al_fitr_day("Ramadan Feast")
        self._add_eid_al_fitr_day_two("Ramadan Feast Holiday")
        self._add_eid_al_fitr_day_three("Ramadan Feast Holiday")

        # Sacrifice Feast.
        # Date of observance is announced yearly, This is an estimate.
        self._add_eid_al_adha_day("Sacrifice Feast")
        self._add_eid_al_adha_day_two("Sacrifice Feast Holiday")
        self._add_eid_al_adha_day_three("Sacrifice Feast Holiday")
        self._add_eid_al_adha_day_four("Sacrifice Feast Holiday")


class TR(Turkey):
    pass


class TUR(Turkey):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import JAN, MAR, MAY, JUL, OCT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Malawi(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://www.officeholidays.com/countries/malawi
    https://www.timeanddate.com/holidays/malawi/
    """

    country = "MW"

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date, days: int = +1) -> None:
        if self.observed and self._is_weekend(dt):
            self._add_holiday(
                "%s (Observed)" % self[dt], dt + td(+2 if self._is_saturday(dt) else days)
            )

    def _populate(self, year):
        # Observed since 2000
        if year <= 1999:
            return None

        super()._populate(year)

        self._add_observed(self._add_new_years_day("New Year's Day"))

        self._add_observed(self._add_holiday("John Chilembwe Day", JAN, 15))

        self._add_observed(self._add_holiday("Martyrs Day", MAR, 3))

        self._add_good_friday("Good Friday")

        self._add_easter_monday("Easter Monday")

        self._add_observed(self._add_labor_day("Labour Day"))

        self._add_observed(self._add_holiday("Kamuzu Day", MAY, 14))

        self._add_observed(self._add_holiday("Independence Day", JUL, 6))

        self._add_observed(self._add_holiday("Mother's Day", OCT, 15))

        self._add_observed(self._add_christmas_day("Christmas Day"), days=+2)

        self._add_observed(self._add_christmas_day_two("Boxing Day"), days=+2)


class MW(Malawi):
    pass


class MWI(Malawi):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

__all__ = ("DateLike", "HolidayBase", "HolidaySum")

import copy
import warnings
from calendar import isleap
from datetime import date, datetime, timedelta, timezone
from gettext import NullTranslations, gettext, translation
from pathlib import Path
from typing import Any, Dict, Iterable, List, Mapping, Optional, Set, Tuple, Union, cast

from dateutil.parser import parse

from holidays.calendars.gregorian import MON, TUE, WED, THU, FRI, SAT, SUN
from holidays.constants import HOLIDAY_NAME_DELIMITER, ALL_CATEGORIES, PUBLIC
from holidays.helpers import _normalize_tuple

DateArg = Union[date, Tuple[int, int]]
DateLike = Union[date, datetime, str, float, int]
SpecialHoliday = Union[Tuple[int, int, str], Tuple[Tuple[int, int, str], ...]]

gettext = gettext


class HolidayBase(Dict[date, str]):
    """
    A dict-like object containing the holidays for a specific country (and
    province or state if so initiated); inherits the dict class (so behaves
    similarly to a dict). Dates without a key in the Holiday object are not
    holidays.

    The key of the object is the date of the holiday and the value is the name
    of the holiday itself. When passing the date as a key, the date can be
    expressed as one of the following formats:

    * datetime.datetime type;
    * datetime.date types;
    * a float representing a Unix timestamp;
    * or a string of any format (recognized by datetime.parse).

    The key is always returned as a `datetime.date` object.

    To maximize speed, the list of holidays is built as needed on the fly, one
    calendar year at a time. When you instantiate the object, it is empty, but
    the moment a key is accessed it will build that entire year's list of
    holidays. To pre-populate holidays, instantiate the class with the years
    argument:

    us_holidays = holidays.US(years=2020)

    It is generally instantiated using the :func:`country_holidays` function.

    The key of the :class:`dict`-like :class:`HolidayBase` object is the
    `date` of the holiday, and the value is the name of the holiday itself.
    Dates where a key is not present are not public holidays (or, if
    **observed** is False, days when a public holiday is observed).

    When passing the `date` as a key, the `date` can be expressed in one of the
    following types:

    * :class:`datetime.date`,
    * :class:`datetime.datetime`,
    * a :class:`str` of any format recognized by :func:`dateutil.parser.parse`,
    * or a :class:`float` or :class:`int` representing a POSIX timestamp.

    The key is always returned as a :class:`datetime.date` object.

    To maximize speed, the list of public holidays is built on the fly as
    needed, one calendar year at a time. When the object is instantiated
    without a **years** parameter, it is empty, but, unless **expand** is set
    to False, as soon as a key is accessed the class will calculate that entire
    year's list of holidays and set the keys with them.

    If you need to list the holidays as opposed to querying individual dates,
    instantiate the class with the **years** parameter.

    Example usage:

    >>> from holidays import country_holidays
    >>> us_holidays = country_holidays('US')
    # For a specific subdivisions (e.g. state or province):
    >>> california_holidays = country_holidays('US', subdiv='CA')

    The below will cause 2015 holidays to be calculated on the fly:

    >>> from datetime import date
    >>> assert date(2015, 1, 1) in us_holidays

    This will be faster because 2015 holidays are already calculated:

    >>> assert date(2015, 1, 2) not in us_holidays

    The :class:`HolidayBase` class also recognizes strings of many formats
    and numbers representing a POSIX timestamp:

    >>> assert '2014-01-01' in us_holidays
    >>> assert '1/1/2014' in us_holidays
    >>> assert 1388597445 in us_holidays

    Show the holiday's name:

    >>> us_holidays.get('2014-01-01')
    "New Year's Day"

    Check a range:

    >>> us_holidays['2014-01-01': '2014-01-03']
    [datetime.date(2014, 1, 1)]

    List all 2020 holidays:

    >>> us_holidays = country_holidays('US', years=2020)
    >>> for day in us_holidays.items():
    ...     print(day)
    (datetime.date(2020, 1, 1), "New Year's Day")
    (datetime.date(2020, 1, 20), 'Martin Luther King Jr. Day')
    (datetime.date(2020, 2, 17), "Washington's Birthday")
    (datetime.date(2020, 5, 25), 'Memorial Day')
    (datetime.date(2020, 7, 4), 'Independence Day')
    (datetime.date(2020, 7, 3), 'Independence Day (Observed)')
    (datetime.date(2020, 9, 7), 'Labor Day')
    (datetime.date(2020, 10, 12), 'Columbus Day')
    (datetime.date(2020, 11, 11), 'Veterans Day')
    (datetime.date(2020, 11, 26), 'Thanksgiving')
    (datetime.date(2020, 12, 25), 'Christmas Day')

    Some holidays are only present in parts of a country:

    >>> us_pr_holidays = country_holidays('US', subdiv='PR')
    >>> assert '2018-01-06' not in us_holidays
    >>> assert '2018-01-06' in us_pr_holidays

    Append custom holiday dates by passing one of:

    * a :class:`dict` with date/name key/value pairs (e.g.
      ``{'2010-07-10': 'My birthday!'}``),
    * a list of dates (as a :class:`datetime.date`, :class:`datetime.datetime`,
      :class:`str`, :class:`int`, or :class:`float`); ``'Holiday'`` will be
      used as a description,
    * or a single date item (of one of the types above); ``'Holiday'`` will be
      used as a description:

    >>> custom_holidays = country_holidays('US', years=2015)
    >>> custom_holidays.update({'2015-01-01': "New Year's Day"})
    >>> custom_holidays.update(['2015-07-01', '07/04/2015'])
    >>> custom_holidays.update(date(2015, 12, 25))
    >>> assert date(2015, 1, 1) in custom_holidays
    >>> assert date(2015, 1, 2) not in custom_holidays
    >>> assert '12/25/2015' in custom_holidays

    For special (one-off) country-wide holidays handling use
    :attr:`special_holidays`:

    .. code-block:: python

        special_holidays = {
            1977: ((JUN, 7, "Silver Jubilee of Elizabeth II"),),
            1981: ((JUL, 29, "Wedding of Charles and Diana"),),
            1999: ((DEC, 31, "Millennium Celebrations"),),
            2002: ((JUN, 3, "Golden Jubilee of Elizabeth II"),),
            2011: ((APR, 29, "Wedding of William and Catherine"),),
            2012: ((JUN, 5, "Diamond Jubilee of Elizabeth II"),),
            2022: (
                (JUN, 3, "Platinum Jubilee of Elizabeth II"),
                (SEP, 19, "State Funeral of Queen Elizabeth II"),
            ),
        }

        def _populate(self, year):
            super()._populate(year)

            ...

    For more complex logic, like 4th Monday of January, you can inherit the
    :class:`HolidayBase` class and define your own :meth:`_populate` method.
    See documentation for examples.
    """

    country: str
    """The country's ISO 3166-1 alpha-2 code."""
    market: str
    """The market's ISO 3166-1 alpha-2 code."""
    subdivisions: Tuple[str, ...] = ()
    """The subdivisions supported for this country (see documentation)."""
    years: Set[int]
    """The years calculated."""
    expand: bool
    """Whether the entire year is calculated when one date from that year
    is requested."""
    observed: bool
    """Whether dates when public holiday are observed are included."""
    subdiv: Optional[str] = None
    """The subdiv requested."""
    special_holidays: Dict[int, SpecialHoliday] = {}
    """A list of the country-wide special (as opposite to regular) holidays for
    a specific year."""
    _deprecated_subdivisions: Tuple[str, ...] = ()
    """Other subdivisions whose names are deprecated or aliases of the official
    ones."""
    weekend: Set[int] = {SAT, SUN}
    """Country weekend days."""
    default_language: Optional[str] = None
    """The entity language used by default."""
    categories: Optional[Set[str]] = None
    """Requested holiday categories."""
    supported_categories: Set[str] = set()
    """All holiday categories supported by this entity."""
    supported_languages: Tuple[str, ...] = ()
    """All languages supported by this entity."""

    def __init__(
        self,
        years: Optional[Union[int, Iterable[int]]] = None,
        expand: bool = True,
        observed: bool = True,
        subdiv: Optional[str] = None,
        prov: Optional[str] = None,  # Deprecated.
        state: Optional[str] = None,  # Deprecated.
        language: Optional[str] = None,
        categories: Optional[Tuple[str]] = None,
    ) -> None:
        """
        :param years:
            The year(s) to pre-calculate public holidays for at instantiation.

        :param expand:
            Whether the entire year is calculated when one date from that year
            is requested.

        :param observed:
            Whether to include the dates when public holiday are observed
            (e.g. a holiday falling on a Sunday being observed the
            following Monday). This doesn't work for all countries.

        :param subdiv:
            The subdivision (e.g. state or province); not implemented for all
            countries (see documentation).

        :param prov:
            *deprecated* use subdiv instead.

        :param state:
            *deprecated* use subdiv instead.

        :param language:
            The language which the returned holiday names will be translated
            into. It must be an ISO 639-1 (2-letter) language code. If the
            language translation is not supported the original holiday names
            will be used.

        :param categories:
            Requested holiday categories.

        :return:
            A :class:`HolidayBase` object matching the **country**.
        """
        super().__init__()

        self.expand = expand
        self.language = language.lower() if language else None
        self.observed = observed
        self.subdiv = subdiv or prov or state
        self.categories = set(categories) if categories else {PUBLIC}

        self.tr = gettext  # Default translation method.

        if prov or state:
            warnings.warn(
                f"Arguments prov and state are deprecated, use subdiv='{prov or state}' instead.",
                DeprecationWarning,
            )

        if not isinstance(self, HolidaySum):
            if subdiv and subdiv not in set(self.subdivisions + self._deprecated_subdivisions):
                if hasattr(self, "market"):
                    error = f"Market '{self.market}' does not have subdivision " f"'{subdiv}'"
                else:
                    error = f"Country '{self.country}' does not have subdivision " f"'{subdiv}'"
                raise NotImplementedError(error)

            if subdiv and subdiv in self._deprecated_subdivisions:
                warnings.warn(
                    "This subdivision is deprecated and will be removed after "
                    "Dec, 1 2023. The list of supported subdivisions: "
                    f"{', '.join(sorted(self.subdivisions))}.",
                    DeprecationWarning,
                )

            unknown_categories = self.categories.difference(ALL_CATEGORIES)
            if len(unknown_categories) > 0:
                raise NotImplementedError(
                    f"Category is not supported: {', '.join(unknown_categories)}."
                )

            name = getattr(self, "country", getattr(self, "market", None))
            if name:
                locale_path = Path(__file__).with_name("locale")
                translator: NullTranslations
                translations = {
                    # Collect `language` part from
                    # holidays/locale/<language>/LC_MESSAGES/country.po
                    translation.parts[-3]
                    for translation in locale_path.rglob(f"{name}.mo")
                }
                if language and language in translations:
                    translator = translation(
                        name, languages=[language], localedir=str(locale_path)
                    )
                else:
                    translator = translation(name, fallback=True, localedir=str(locale_path))
                self.tr = translator.gettext

        if isinstance(years, int):
            self.years = {years}
        else:
            self.years = set(years) if years is not None else set()

        for year in self.years:
            self._populate(year)

    def __add__(self, other: Union[int, "HolidayBase", "HolidaySum"]) -> "HolidayBase":
        """Add another dictionary of public holidays creating a
        :class:`HolidaySum` object.

        :param other:
            The dictionary of public holiday to be added.

        :return:
            A :class:`HolidaySum` object unless the other object cannot be
            added, then :class:`self`.
        """
        if isinstance(other, int) and other == 0:
            # Required to sum() list of holidays
            # sum([h1, h2]) is equivalent to (0 + h1 + h2).
            return self

        if not isinstance(other, (HolidayBase, HolidaySum)):
            raise TypeError("Holiday objects can only be added with other Holiday objects")

        return HolidaySum(self, other)

    def __bool__(self) -> bool:
        return len(self) > 0

    def __contains__(self, key: object) -> bool:
        """Return true if date is in self, false otherwise. Accepts a date in
        the following types:

        * :class:`datetime.date`,
        * :class:`datetime.datetime`,
        * a :class:`str` of any format recognized by
          :func:`dateutil.parser.parse`,
        * or a :class:`float` or :class:`int` representing a POSIX timestamp.
        """

        if not isinstance(key, (date, datetime, float, int, str)):
            raise TypeError(f"Cannot convert type '{type(key)}' to date.")

        return dict.__contains__(cast("Mapping[Any, Any]", self), self.__keytransform__(key))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HolidayBase):
            return False

        for attribute_name in self.__attribute_names:
            if getattr(self, attribute_name, None) != getattr(other, attribute_name, None):
                return False

        return dict.__eq__(self, other)

    def __getitem__(self, key: DateLike) -> Any:
        if isinstance(key, slice):
            if not key.start or not key.stop:
                raise ValueError("Both start and stop must be given.")

            start = self.__keytransform__(key.start)
            stop = self.__keytransform__(key.stop)

            if key.step is None:
                step = 1
            elif isinstance(key.step, timedelta):
                step = key.step.days
            elif isinstance(key.step, int):
                step = key.step
            else:
                raise TypeError(f"Cannot convert type '{type(key.step)}' to int.")

            if step == 0:
                raise ValueError("Step value must not be zero.")

            date_diff = stop - start
            if date_diff.days < 0 <= step or date_diff.days >= 0 > step:
                step *= -1

            days_in_range = []
            for delta_days in range(0, date_diff.days, step):
                day = start + timedelta(days=delta_days)
                if day in self:
                    days_in_range.append(day)

            return days_in_range

        return dict.__getitem__(self, self.__keytransform__(key))

    def __keytransform__(self, key: DateLike) -> date:
        """Transforms the date from one of the following types:

        * :class:`datetime.date`,
        * :class:`datetime.datetime`,
        * a :class:`str` of any format recognized by
          :func:`dateutil.parser.parse`,
        * or a :class:`float` or :class:`int` representing a POSIX timestamp

        to :class:`datetime.date`, which is how it's stored by the class."""

        # Try to catch `date` and `str` type keys first.
        if type(key) == date:  # Key has `date` type.
            dt = key

        elif type(key) == str:  # Key has `str` type.
            try:
                dt = parse(key).date()
            except (OverflowError, ValueError):
                raise ValueError(f"Cannot parse date from string '{key}'")

        # Check all other types.
        elif isinstance(key, datetime):  # Key type is derived from `datetime`.
            dt = key.date()

        # Must go after the `isinstance(key, datetime)` check
        # as datetime is derived from `date`.
        elif isinstance(key, date):  # Key type is derived from `date`.
            dt = key

        elif isinstance(key, (float, int)):  # Key type is derived from `float` or `int`.
            dt = datetime.fromtimestamp(key, timezone.utc).date()

        else:  # Key type is not supported.
            raise TypeError(f"Cannot convert type '{type(key)}' to date.")

        # Automatically expand for `expand=True` cases.
        if self.expand and dt.year not in self.years:
            self.years.add(dt.year)
            self._populate(dt.year)

        return dt

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, HolidayBase):
            return True

        for attribute_name in self.__attribute_names:
            if getattr(self, attribute_name, None) != getattr(other, attribute_name, None):
                return True

        return dict.__ne__(self, other)

    def __radd__(self, other: Any) -> "HolidayBase":
        return self.__add__(other)

    def __reduce__(self) -> Union[str, Tuple[Any, ...]]:
        return super().__reduce__()

    def __repr__(self) -> str:
        if self:
            return super().__repr__()

        parts = []
        if hasattr(self, "market"):
            parts.append(f"holidays.financial_holidays({self.market!r}")
        elif hasattr(self, "country"):
            if parts:
                parts.append(" + ")
            parts.append(f"holidays.country_holidays({self.country!r}")
            if self.subdiv:
                parts.append(f", subdiv={self.subdiv!r}")
        parts.append(")")

        return "".join(parts)

    def __setattr__(self, key: str, value: Any) -> None:
        dict.__setattr__(self, key, value)

        if self and key in {"categories", "observed"}:
            self.clear()
            for year in self.years:  # Re-populate holidays for each year.
                self._populate(year)

    def __setitem__(self, key: DateLike, value: str) -> None:
        if key in self:
            # If there are multiple holidays on the same date
            # order their names alphabetically.
            holiday_names = set(self[key].split(HOLIDAY_NAME_DELIMITER))
            holiday_names.add(value)
            value = HOLIDAY_NAME_DELIMITER.join(sorted(holiday_names))

        dict.__setitem__(self, self.__keytransform__(key), value)

    def __str__(self) -> str:
        if self:
            return super().__str__()

        parts = []
        for attribute_name in self.__attribute_names:
            parts.append("'%s': %s" % (attribute_name, getattr(self, attribute_name, None)))

        return f"{{{', '.join(parts)}}}"

    @property
    def __attribute_names(self):
        return (
            "country",
            "expand",
            "language",
            "market",
            "observed",
            "subdiv",
            "years",
        )

    @staticmethod
    def _is_leap_year(year: int) -> bool:
        """
        Returns True if the year is leap. Returns False otherwise.
        """
        return isleap(year)

    def _add_holiday(self, *args) -> Optional[date]:
        """Add a holiday.

        This method accepts either `name: str, dt: date` or
        `name: str, month: int, day: int` arguments.
        """
        name, dt = self._parse_holiday(*args)
        if dt.year != self._year:
            return None

        self[dt] = self.tr(name)
        return dt

    def _add_subdiv_holidays(self):
        """Populate subdivision holidays."""
        if self.subdiv is not None:
            subdiv = self.subdiv.replace("-", "_").replace(" ", "_").lower()
            add_subdiv_holidays = getattr(self, f"_add_subdiv_{subdiv}_holidays", None)
            if add_subdiv_holidays and callable(add_subdiv_holidays):
                add_subdiv_holidays()

    def _check_weekday(self, weekday: int, *args) -> bool:
        """
        Returns True if `weekday` equals to the date's week day.
        Returns False otherwise.
        """
        dt = args[0] if len(args) == 1 else date(self._year, *args)
        return dt.weekday() == weekday

    def _get_nth_weekday_from(self, n: int, weekday: int, *args) -> date:
        """
        Return date of a n-th weekday after (n is positive)
        or before (n is negative) a specific date
        (e.g. 1st Monday, 2nd Saturday, etc).
        """
        from_dt = args[0] if len(args) == 1 else date(self._year, *args)
        if n > 0:
            delta = (n - 1) * 7 + (weekday - from_dt.weekday()) % 7
        else:
            delta = (n + 1) * 7 - (from_dt.weekday() - weekday) % 7
        return from_dt + timedelta(days=delta)

    def _get_nth_weekday_of_month(self, n: int, weekday: int, month: int) -> date:
        """
        Return date of n-th weekday of month for current year
        (e.g. 1st Monday of Apr, 2nd Friday of June, etc).
        If n is negative the countdown starts at the end of month
        (i.e. -1 is last).
        """
        year = self._year
        if n < 0:
            month += 1
            if month > 12:
                month = 1
                year += 1
            start_date = date(year, month, 1) + timedelta(days=-1)
        else:
            start_date = date(year, month, 1)
        return self._get_nth_weekday_from(n, weekday, start_date)

    def _is_friday(self, *args) -> bool:
        return self._check_weekday(FRI, *args)

    def _is_monday(self, *args) -> bool:
        return self._check_weekday(MON, *args)

    def _is_saturday(self, *args) -> bool:
        return self._check_weekday(SAT, *args)

    def _is_sunday(self, *args) -> bool:
        return self._check_weekday(SUN, *args)

    def _is_thursday(self, *args) -> bool:
        return self._check_weekday(THU, *args)

    def _is_tuesday(self, *args) -> bool:
        return self._check_weekday(TUE, *args)

    def _is_wednesday(self, *args) -> bool:
        return self._check_weekday(WED, *args)

    def _is_weekend(self, *args):
        """
        Returns True if date's week day is a weekend day.
        Returns False otherwise.
        """
        dt = args[0] if len(args) == 1 else date(self._year, *args)

        return dt.weekday() in self.weekend

    def _parse_holiday(self, *args) -> Tuple[str, date]:
        """Parse holiday data."""
        if len(args) == 2:
            name, dt = args
            if not isinstance(dt, date):
                raise TypeError(
                    "Invalid argument type: expected <class 'date'> " f"got '{type(dt)}'."
                )
        elif len(args) == 3:
            name, month, day = args
            dt = date(self._year, month, day)
        else:
            raise TypeError("Incorrect number of arguments.")

        return name, dt

    def _populate(self, year: int) -> None:
        """This is a private method that populates (generates and adds) holidays
        for a given year. To keep things fast, it assumes that no holidays for
        the year have already been populated. It is required to be called
        internally by any country populate() method, while should not be called
        directly from outside.
        To add holidays to an object, use the update() method.

        :param year:
            The year to populate with holidays.

        >>> from holidays import country_holidays
        >>> us_holidays = country_holidays('US', years=2020)
        # to add new holidays to the object:
        >>> us_holidays.update(country_holidays('US', years=2021))
        """

        self._year = year

        # Populate items from the special holidays list.
        for month, day, name in _normalize_tuple(self.special_holidays.get(year, ())):
            self._add_holiday(name, month, day)

        # Populate categories holidays.
        self._populate_categories()

        # Populate subdivision holidays.
        self._add_subdiv_holidays()

    def _populate_categories(self):
        for category in self.categories:
            # Populate items from the special holidays list for all categories.
            special_category_holidays = getattr(self, f"special_{category}_holidays", None)
            if special_category_holidays:
                for month, day, name in _normalize_tuple(
                    special_category_holidays.get(self._year, ())
                ):
                    self._add_holiday(name, month, day)

            populate_category_holidays = getattr(self, f"_populate_{category}_holidays", None)
            if populate_category_holidays and callable(populate_category_holidays):
                populate_category_holidays()

    def append(self, *args: Union[Dict[DateLike, str], List[DateLike], DateLike]) -> None:
        """Alias for :meth:`update` to mimic list type."""
        return self.update(*args)

    def copy(self):
        """Return a copy of the object."""
        return copy.copy(self)

    def get(self, key: DateLike, default: Union[str, Any] = None) -> Union[str, Any]:
        """Return the holiday name for a date if date is a holiday, else
        default. If default is not given, it defaults to None, so that this
        method never raises a KeyError. If more than one holiday is present,
        they are separated by a comma.

        :param key:
            The date expressed in one of the following types:

            * :class:`datetime.date`,
            * :class:`datetime.datetime`,
            * a :class:`str` of any format recognized by
              :func:`dateutil.parser.parse`,
            * or a :class:`float` or :class:`int` representing a POSIX
              timestamp.

        :param default:
            The default value to return if no value is found.
        """
        return dict.get(self, self.__keytransform__(key), default)

    def get_list(self, key: DateLike) -> List[str]:
        """Return a list of all holiday names for a date if date is a holiday,
        else empty string.

        :param key:
            The date expressed in one of the following types:

            * :class:`datetime.date`,
            * :class:`datetime.datetime`,
            * a :class:`str` of any format recognized by
              :func:`dateutil.parser.parse`,
            * or a :class:`float` or :class:`int` representing a POSIX
              timestamp.
        """
        return [name for name in self.get(key, "").split(HOLIDAY_NAME_DELIMITER) if name]

    def get_named(
        self, holiday_name: str, lookup="icontains", split_multiple_names=True
    ) -> List[date]:
        """Return a list of all holiday dates matching the provided holiday
        name. The match will be made case insensitively and partial matches
        will be included by default.

        :param holiday_name:
            The holiday's name to try to match.
        :param lookup:
            The holiday name lookup type:
                contains - case sensitive contains match;
                exact - case sensitive exact match;
                startswith - case sensitive starts with match;
                icontains - case insensitive contains match;
                iexact - case insensitive exact match;
                istartswith - case insensitive starts with match;
        :param split_multiple_names:
            Either use the exact name for each date or split it by holiday
            name delimiter.

        :return:
            A list of all holiday dates matching the provided holiday name.
        """
        holiday_name_dates = (
            ((k, name) for k, v in self.items() for name in v.split(HOLIDAY_NAME_DELIMITER))
            if split_multiple_names
            else ((k, v) for k, v in self.items())
        )

        if lookup == "icontains":
            holiday_name_lower = holiday_name.lower()
            return [dt for dt, name in holiday_name_dates if holiday_name_lower in name.lower()]
        elif lookup == "exact":
            return [dt for dt, name in holiday_name_dates if holiday_name == name]
        elif lookup == "contains":
            return [dt for dt, name in holiday_name_dates if holiday_name in name]
        elif lookup == "startswith":
            return [
                dt for dt, name in holiday_name_dates if holiday_name == name[: len(holiday_name)]
            ]
        elif lookup == "iexact":
            holiday_name_lower = holiday_name.lower()
            return [dt for dt, name in holiday_name_dates if holiday_name_lower == name.lower()]
        elif lookup == "istartswith":
            holiday_name_lower = holiday_name.lower()
            return [
                dt
                for dt, name in holiday_name_dates
                if holiday_name_lower == name[: len(holiday_name)].lower()
            ]

        raise AttributeError(f"Unknown lookup type: {lookup}")

    def pop(self, key: DateLike, default: Union[str, Any] = None) -> Union[str, Any]:
        """If date is a holiday, remove it and return its date, else return
        default.

        :param key:
            The date expressed in one of the following types:

            * :class:`datetime.date`,
            * :class:`datetime.datetime`,
            * a :class:`str` of any format recognized by
              :func:`dateutil.parser.parse`,
            * or a :class:`float` or :class:`int` representing a POSIX
              timestamp.

        :param default:
            The default value to return if no match is found.

        :return:
            The date removed.

        :raise:
            KeyError if date is not a holiday and default is not given.
        """
        if default is None:
            return dict.pop(self, self.__keytransform__(key))

        return dict.pop(self, self.__keytransform__(key), default)

    def pop_named(self, name: str) -> List[date]:
        """Remove (no longer treat at as holiday) all dates matching the
        provided holiday name. The match will be made case insensitively and
        partial matches will be removed.

        :param name:
            The holiday's name to try to match.

        :param default:
            The default value to return if no match is found.

        :return:
            A list of dates removed.

        :raise:
            KeyError if date is not a holiday and default is not given.
        """
        use_exact_name = HOLIDAY_NAME_DELIMITER in name
        dts = self.get_named(name, split_multiple_names=not use_exact_name)
        if len(dts) == 0:
            raise KeyError(name)

        popped = []
        for dt in dts:
            holiday_names = self[dt].split(HOLIDAY_NAME_DELIMITER)
            self.pop(dt)
            popped.append(dt)

            # Keep the rest of holidays falling on the same date.
            if not use_exact_name:
                name_lower = name.lower()
                holiday_names = [
                    holiday_name
                    for holiday_name in holiday_names
                    if name_lower not in holiday_name.lower()
                ]

                if len(holiday_names) > 0:
                    self[dt] = HOLIDAY_NAME_DELIMITER.join(holiday_names)

        return popped

    def update(  # type: ignore[override]
        self, *args: Union[Dict[DateLike, str], List[DateLike], DateLike]
    ) -> None:
        # TODO: fix arguments; should not be *args (cannot properly Type hint)
        """Update the object, overwriting existing dates.

        :param:
            Either another dictionary object where keys are dates and values
            are holiday names, or a single date (or a list of dates) for which
            the value will be set to "Holiday".

            Dates can be expressed in one or more of the following types:

            * :class:`datetime.date`,
            * :class:`datetime.datetime`,
            * a :class:`str` of any format recognized by
              :func:`dateutil.parser.parse`,
            * or a :class:`float` or :class:`int` representing a POSIX
              timestamp.
        """
        for arg in args:
            if isinstance(arg, dict):
                for key, value in arg.items():
                    self[key] = value
            elif isinstance(arg, list):
                for item in arg:
                    self[item] = "Holiday"
            else:
                self[arg] = "Holiday"


class HolidaySum(HolidayBase):
    """
    Returns a :class:`dict`-like object resulting from the addition of two or
    more individual dictionaries of public holidays. The original dictionaries
    are available as a :class:`list` in the attribute :attr:`holidays,` and
    :attr:`country` and :attr:`subdiv` attributes are added
    together and could become :class:`list` s. Holiday names, when different,
    are merged. All years are calculated (expanded) for all operands.
    """

    country: Union[str, List[str]]  # type: ignore[assignment]
    """Countries included in the addition."""
    market: Union[str, List[str]]  # type: ignore[assignment]
    """Markets included in the addition."""
    subdiv: Optional[Union[str, List[str]]]  # type: ignore[assignment]
    """Subdivisions included in the addition."""
    holidays: List[HolidayBase]
    """The original HolidayBase objects included in the addition."""
    years: Set[int]
    """The years calculated."""

    def __init__(
        self, h1: Union[HolidayBase, "HolidaySum"], h2: Union[HolidayBase, "HolidaySum"]
    ) -> None:
        """
        :param h1:
            The first HolidayBase object to add.

        :param h2:
            The other HolidayBase object to add.

        Example:

        >>> from holidays import country_holidays
        >>> nafta_holidays = country_holidays('US', years=2020) + \
country_holidays('CA') + country_holidays('MX')
        >>> dates = sorted(nafta_holidays.items(), key=lambda x: x[0])
        >>> from pprint import pprint
        >>> pprint(dates[:10], width=72)
        [(datetime.date(2020, 1, 1), "Año Nuevo"),
         (datetime.date(2020, 1, 20), 'Martin Luther King Jr. Day'),
         (datetime.date(2020, 2, 3),
          'Día de la Constitución'),
         (datetime.date(2020, 2, 17), "Washington's Birthday, Family Day"),
         (datetime.date(2020, 3, 16),
          "Natalicio de Benito Juárez"),
         (datetime.date(2020, 4, 10), 'Good Friday'),
         (datetime.date(2020, 5, 1), 'Día del Trabajo'),
         (datetime.date(2020, 5, 18), 'Victoria Day')]
        """
        # Store originals in the holidays attribute.
        self.holidays = []
        for operand in (h1, h2):
            if isinstance(operand, HolidaySum):
                self.holidays.extend(operand.holidays)
            else:
                self.holidays.append(operand)

        kwargs: Dict[str, Any] = {}
        # Join years, expand and observed.
        kwargs["years"] = h1.years | h2.years
        kwargs["expand"] = h1.expand or h2.expand
        kwargs["observed"] = h1.observed or h2.observed
        # Join country and subdivisions data.
        # TODO: this way makes no sense: joining Italy Catania (IT, CA) with
        # USA Mississippi (US, MS) and USA Michigan (US, MI) yields
        # country=["IT", "US"] and subdiv=["CA", "MS", "MI"], which could very
        # well be California and Messina and Milano, or Catania, Mississippi
        # and Milano, or ... you get the picture.
        # Same goes when countries and markets are being mixed (working, yet
        # still nonsensical).
        for attr in ("country", "market", "subdiv"):
            if (
                getattr(h1, attr, None)
                and getattr(h2, attr, None)
                and getattr(h1, attr) != getattr(h2, attr)
            ):
                a1 = (
                    getattr(h1, attr)
                    if isinstance(getattr(h1, attr), list)
                    else [getattr(h1, attr)]
                )
                a2 = (
                    getattr(h2, attr)
                    if isinstance(getattr(h2, attr), list)
                    else [getattr(h2, attr)]
                )
                value = a1 + a2
            else:
                value = getattr(h1, attr, None) or getattr(h2, attr, None)

            if attr == "subdiv":
                kwargs[attr] = value
            else:
                setattr(self, attr, value)

        HolidayBase.__init__(self, **kwargs)

    def _populate(self, year):
        for operand in self.holidays:
            operand._populate(year)
            self.update(cast("Dict[DateLike, str]", operand))
